@@ -2,29 +2,34 @@
 # This tests the various methods of the Graphics::Framebuffer module
 # and shows the only way threads will work with it.
 
+## THIS SHOWS HOW TO USE WITH MALI GPUs
+
 use strict;
 use Switch;
 
 use threads;
+use threads::shared;
 use Graphics::Framebuffer;
 use Sys::CPU;
-use Data::Dumper::Simple;
 
+my $mali = 0;
+my $hold : shared = 0;
+my @waiting : shared;
 my $Threads = $ARGV[0] || Sys::CPU::cpu_count();
+$mali = 1 if ($ARGV[1] =~ /mali/i);
 print "\nRunning on a ",Sys::CPU::cpu_type(), " with $Threads threads\n";
 sleep 3;
 my @framebuffer;
 foreach my $thr (0..$Threads) {
-    $framebuffer[$thr] = Graphics::Framebuffer->new('FB_DEVICE' => '/dev/fb0','FILE_MODE' => 0);
+    $framebuffer[$thr] = Graphics::Framebuffer->new('FB_DEVICE' => '/dev/fb0','FILE_MODE' => 0,'MALI' => $mali);
 }
-my $screen_width  = $framebuffer[0]->{'XRES'};
-my $screen_height = $framebuffer[0]->{'YRES'};
-my $cols = $Threads / 2;
+my ($screen_width,$screen_height)  = $framebuffer[0]->screen_dimensions();
 
 
 $framebuffer[0]->cls();
-my $cls = int(rand($screen_width));
+
 foreach my $page (1..$Threads) {
+    $waiting[$page] = 0;
     threads->create(
         sub {
             my $Page = shift;
@@ -34,14 +39,27 @@ foreach my $page (1..$Threads) {
             }
             while(1) {
                 attract($Page,$Cols);
+                while($hold) {
+                    threads->yield();
+                    $waiting[$Page] = 0;
+                    sleep 1;
+                }
+                $waiting[$Page] = 1;
             }
         },
-        $page,$cols
+        $page
     );
 }
 while(1) {
-    threads->yield();
-    sleep 1;
+    $hold = 0;
+    sleep 30;
+    $hold = 1;
+    foreach my $page (1..$Threads) {
+        while($waiting[$page]) {
+            sleep 1;
+        }
+    }
+    $framebuffer[0]->cls();
 }
 
 
@@ -54,24 +72,22 @@ while(1) {
 
 sub attract {
     my $page = shift;
-    my $cols = 4; # shift;
     my $red  = int(rand(256));
     my $grn  = int(rand(256));
     my $blu  = int(rand(256));
     my $x    = int(rand($screen_width));
     my $y    = int(rand($screen_height));
-    my $w    = int(rand($screen_width/$cols));
-    my $h    = int(rand($screen_height/$cols));
-    my $rx   = int(rand($screen_width/$cols));
-    my $ry   = int(rand($screen_height/$cols));
+    my $w    = int(rand($screen_width/3));
+    my $h    = int(rand($screen_height/3));
+    my $rx   = int(rand($screen_width/5));
+    my $ry   = int(rand($screen_height/5));
     my $sd   = int(rand(360));
     my $ed   = int(rand(360));
     my $gr   = (rand(100)/100);
-    my $mode = int(rand(5));
+    my $mode = int(rand(4));
     my $type = int(rand(7));
     my $size = int(rand(7)) - 3;
     my $arc  = int(rand(3));
-    $framebuffer[$page]->cls() if ($x >= ($cls - 4) && $x <= ($cls + 4));
 
     $framebuffer[$page]->set_color({'red' => $red,'green' => $grn,'blue' => $blu});
     $framebuffer[$page]->draw_mode($mode);
@@ -104,7 +120,6 @@ sub attract {
             $framebuffer[$page]->polygon({'pixel_size' => $size,'coordinates' => \@poly, 'filled' => (int(rand(2)))});
         }
         else { # This can be a large memory hog.  Disabled (random number doesn't go this high) for the moment.
-#            $framebuffer[$Page]->clip_set({'x' => $x-50,'y' => $y-50,'xx' => $x+50,'yy' => $y+50});
             $framebuffer[$page]->fill({'x' => $x,'y' => $y});
         }
     }
