@@ -7,7 +7,9 @@ use Graphics::Framebuffer;
 use Term::ANSIScreen qw(:cursor);
 
 $| = 1;
-my $F = Graphics::Framebuffer->new('FB_DEVICE' => '/dev/fb0','FILE_MODE' => 0);
+my $mali = 0;
+$mali = 1 if ($ARGV[0] =~ /mali/i);
+my $F = Graphics::Framebuffer->new('MALI' => $mali, 'FB_DEVICE' => '/dev/fb0','FILE_MODE' => 1);
 
 my $screen_width  = $F->{'XRES'};
 my $screen_height = $F->{'YRES'};
