@@ -83,8 +83,8 @@ use constant {
 };
 use 5.010;
 
-use Switch;          # Yes, a touch of new Perl. Switch is so much nicer
-                     # than a ton of if-else's
+use Switch;                # Yes, a touch of new Perl. Switch is so much nicer
+                           # than a ton of if-else's
 use Math::Trig qw(:pi);    # Yes, yummy PI
 use Sys::Mmap;             # Absolutely necessary to map the screen to a string.
                            # Without this, I doubt this module would exist.
@@ -92,13 +92,11 @@ use Sys::Mmap;             # Absolutely necessary to map the screen to a string.
 use Imager;                # This is used for TrueType font printing, and where
                            # 'Pure Perl' goes up in smoke.
 
-# use Data::Dumper::Simple;
-
 BEGIN {
     require Exporter;
 
     # set the version for version checking
-    our $VERSION   = 4.22;
+    our $VERSION   = 4.23;
     our @ISA       = qw(Exporter);
     our @EXPORT    = qw();
     our @EXPORT_OK = qw();
@@ -125,17 +123,37 @@ my $fb = Graphics::Framebuffer->new(parameter => value);
 
 =over 2
 
-=item B< FB_DEVICE>
+=item B< FB_DEVICE> [/dev/fb0]
 
- Framebuffer device name. Default is /dev/fb0
+ Framebuffer device name.
 
-=item B< FILE_MODE >
+=item B< FILE_MODE > [1 or 0]
 
  Sets the internal drawing system to use file handle mode
  instead of memory mapped mode.  File mode is more stable,
  but a bit slower.  I recommend doing this ONLY if you
  are having issues with memory mapped mode.
 
+=item B< MALI > [1 or 0]
+
+ Forces the drawing engine to ignore some of the data
+ returned by the driver, and do a sensible calculation for
+ drawing parameters.  This is usually required for ARM MALI
+ GPUs.
+
+ It gives a correct calculation of internal variables:
+
+  (Assuming your object variable is $fb)
+
+  $fb->{'BYTES_PER_LINE'}
+  $fb->{'PIXELS'}
+  $fb->{'BYTES'}
+  $fb->{'smem_len'}
+
+ This also means, that if drawing looks funky, then
+ perhaps adjusting these internal variables AFTER the
+ object is created, can fix your issue.
+
 =item B< COLOR_ORDER >
 
 Sets the coloring logic to use the proper color ordering.
@@ -202,21 +220,31 @@ sub new {
         'FBIOGETCMAP'         => 0x4604,
         'FBIOPUTCMAP'         => 0x4605,
         'FBIOPAN_DISPLAY'     => 0x4606,
+        'FBIOGET_CON2FBMAP'   => 0x460F,
+        'FBIOPUT_CON2FBMAP'   => 0x4610,
+        'FBIOBLANK'           => 0x4611,
+        'FBIOGET_GLYPH'       => 0x4615,
+        'FBIOGET_HWCINFO'     => 0x4616,
+        'FBIOPUT_MODEINFO'    => 0x4617,
+        'FBIOGET_DISPINFO'    => 0x4618,
 
         # FLAGS
         'FBINFO_HWACCEL_COPYAREA'  => 0x0100,
         'FBINFO_HWACCEL_FILLRECT'  => 0x0200,
         'FBINFO_HWACCEL_IMAGEBLIT' => 0x0400,
-        'FBINFO_HWACCEL_ROTATE'    => 0x1000,
-        'FBINFO_HWACCEL_XPAN'      => 0x2000,
-        'FBINFO_HWACCEL_YPAN'      => 0x4000,
+        'FBINFO_HWACCEL_ROTATE'    => 0x0800,
+        'FBINFO_HWACCEL_XPAN'      => 0x1000,
+        'FBINFO_HWACCEL_YPAN'      => 0x2000,
+        'FBINFO_HWACCEL_YWRAP'     => 0x4000,
 
+        # I=32,S=16,C=8,A=string
         # Structure Definitions
         'FBioget_vscreeninfo'      => 'I24',
         'FBioget_fscreeninfo'      => 'A16LI4S3ILI2S',
-        'FBinfo_hwaccel_copyarea'  => 'I6',              # dx(32),dy(32),width(32),height(32),sx(32),sy(32)
-        'FBinfo_hwaccel_fillrect'  => 'I6',              # dx(32),dy(32),width(32),height(32),color(32),rop(32)
-        'FBinfo_hwaccel_imageblit' => 'I6C1I2',          # dx(32),dy(32),width(32),height(32),fg_color(32),bg_color(32),depth(8),image pointer(32),color map pointer(32)
+        'FBinfo_hwaccel_fillrect'  => 'L6',              # dx(32),dy(32),width(32),height(32),color(32),rop(32)?
+        'FBinfo_hwaccel_copyarea'  => 'L6',              # dx(32),dy(32),width(32),height(32),sx(32),sy(32)
+        'FBinfo_hwaccel_fillrect'  => 'L6',              # dx(32),dy(32),width(32),height(32),color(32),rop(32)
+        'FBinfo_hwaccel_imageblit' => 'L6C1I1',          # dx(32),dy(32),width(32),height(32),fg_color(32),bg_color(32),depth(8),image pointer(32),color map pointer(32)
                                                          # COLOR MAP:
                                                          #  start(32),length(32),red(16),green(16),blue(16),alpha(16)
 
@@ -257,10 +285,16 @@ sub new {
         $self->{'YOFFSET'}        = $self->{'yoffset'} || 0;
         $self->{'BITS'}           = $self->{'bits_per_pixel'};
         $self->{'BYTES'}          = $self->{'BITS'} / 8;
+        if ($self->{'MALI'}) { # The MALI framebuffer driver is nuts.  This should be a fix
+            $self->{'VXRES'} = $self->{'XRES'};
+            $self->{'VYRES'} = $self->{'YRES'};
+            $self->{'smem_len'}       = $self->{'BYTES'} * ($self->{'XRES'} * $self->{'YRES'});
+        }
         $self->{'PIXELS'}         = (($self->{'XOFFSET'} + $self->{'VXRES'}) * ($self->{'YOFFSET'} + $self->{'VYRES'}));
         $self->{'SIZE'}           = $self->{'PIXELS'} * $self->{'BYTES'};
         $self->{'smem_len'}       = $self->{'BYTES'} * ($self->{'VXRES'} * $self->{'VYRES'}) if (!defined($self->{'smem_len'}) || $self->{'smem_len'} <= 0);
         $self->{'BYTES_PER_LINE'} = int($self->{'smem_len'} / $self->{'VYRES'});
+
 #        print Dumper($self);exit;
 
         bless($self, $class);
@@ -602,8 +636,8 @@ sub plot {
             } ## end unless (exists($self->{'history'...}))
         }
     } ## end else [ if ($size > 1) ]
-    $self->{'X'} = $x;
-    $self->{'Y'} = $y;
+    $self->{'X'} = int($x);
+    $self->{'Y'} = int($y);
 } ## end sub plot
 
 =head2 line
@@ -866,21 +900,21 @@ sub draw_arc {
     my $self   = shift;
     my $params = shift;
 
-    my $x      = $params->{'x'};
-    my $y      = $params->{'y'};
+    my $x      = int($params->{'x'});
+    my $y      = int($params->{'y'});
     my $radius = int($params->{'radius'} || 1);
 
-    my $start_degrees = 0 + sprintf('%.03f', $params->{'start_degrees'}) || 0;
-    my $end_degrees   = 0 + sprintf('%.03f', $params->{'end_degrees'})   || 360;
-    my $granularity   = 0 + sprintf('%.04f', $params->{'granularity'})   || .1;
+    my $start_degrees = $params->{'start_degrees'} || 0;
+    my $end_degrees   = $params->{'end_degrees'}   || 360;
+    my $granularity   = $params->{'granularity'}   || .1;
 
     my $mode = int($params->{'mode'}       || 0);
     my $size = int($params->{'pixel_size'} || 1);
-#    $size = 2 if ($mode == $self->{'PIE'});
     my ($sx, $sy, $degrees, $ox, $oy) = (0, 0, 1, 1, 1);
     $self->{'history'} = {};    # Initialize the history buffer for optimization and xor beautifying.
 
     $degrees = $start_degrees;
+    my $plotted = FALSE;
     if ($start_degrees > $end_degrees) {
         do {
             $sx = int($x - ($radius * sin(($degrees * pi) / 180)));
@@ -888,7 +922,12 @@ sub draw_arc {
             if (($sx <=> $ox) || ($sy <=> $oy)) {
                 switch ($mode) {
                     case ($self->{'ARC'}) {     # Ordinary arc
-                        $self->plot({'x' => $sx, 'y' => $sy, 'pixel_size' => $size});
+                        if ($plotted) { # Fills in the gaps better this way
+                            $self->drawto({'x' => $sx, 'y' => $sy, 'pixel_size' => $size});
+                        } else {
+                            $self->plot({'x' => $sx, 'y' => $sy, 'pixel_size' => $size});
+                            $plotted = TRUE;
+                        }
                     }
                     case ($self->{'PIE'}) {     # Filled arc
                         $self->line({'x' => $x, 'y' => $y, 'xx' => $sx, 'yy' => $sy, 'pixel_size' => $size});
@@ -908,13 +947,19 @@ sub draw_arc {
         } until ($degrees >= 360);
         $degrees = 0;
     } ## end if ($start_degrees > $end_degrees)
+    $plotted = FALSE;
     do {
         $sx = int($x - ($radius * sin(($degrees * pi) / 180)));
         $sy = int($y - ($radius * cos(($degrees * pi) / 180)));
         if (($sx <=> $ox) || ($sy <=> $oy)) {
             switch ($mode) {
                 case ($self->{'ARC'}) {    # Ordinary arc
-                    $self->plot({'x' => $sx, 'y' => $sy, 'pixel_size' => $size});
+                    if ($plotted) { # Fills in the gaps better this way
+                        $self->drawto({'x' => $sx, 'y' => $sy, 'pixel_size' => $size});
+                    } else {
+                        $self->plot({'x' => $sx, 'y' => $sy, 'pixel_size' => $size});
+                        $plotted = TRUE;
+                    }
                 }
                 case ($self->{'PIE'}) {    # Filled arc
                     $self->line({'x' => $x, 'y' => $y, 'xx' => $sx, 'yy' => $sy, 'pixel_size' => $size});
@@ -1090,46 +1135,48 @@ sub ellipse {
     my $self   = shift;
     my $params = shift;
 
-    my $cx      = $params->{'x'};
-    my $cy      = $params->{'y'};
+    my $cx      = int($params->{'x'});
+    my $cy      = int($params->{'y'});
     my $XRadius = int($params->{'xradius'} || 1);
     my $YRadius = int($params->{'yradius'} || 1);
+
     $XRadius = 1 if ($XRadius < 1);
     $YRadius = 1 if ($YRadius < 1);
 
-    my $filled = int($params->{'filled'} || 0);
-    my $fact = $params->{'factor'}     || 1;
-    my $size = $params->{'pixel_size'} || 1;
+    my $filled = int($params->{'filled'}     || 0);
+    my $fact   = $params->{'factor'}         || 1;
+    my $size   = int($params->{'pixel_size'} || 1);
     $size = 1 if ($filled);
 
     my ($old_cyy, $old_cy_y) = (0, 0);
     if ($fact == 0) {    # We don't allow zero values for this
         $fact = 1;
     }
-    my $TwoASquare = (2 * ($XRadius * $XRadius)) * $fact;
-    my $TwoBSquare = (2 * ($YRadius * $YRadius)) * $fact;
-    my $x          = $XRadius;
-    my $y          = 0;
+    my $TwoASquare   = (2 * ($XRadius * $XRadius)) * $fact;
+    my $TwoBSquare   = (2 * ($YRadius * $YRadius)) * $fact;
+    my $x            = $XRadius;
+    my $y            = 0;
     my $XChange      = ($YRadius * $YRadius) * (1 - (2 * $XRadius));
     my $YChange      = ($XRadius * $XRadius);
     my $EllipseError = 0;
     my $StoppingX    = $TwoBSquare * $XRadius;
     my $StoppingY    = 0;
+
     $self->{'history'} = {} unless ($filled || $size > 1);
 
     while ($StoppingX >= $StoppingY) {
-        my $cxx  = $cx + $x;
-        my $cx_x = $cx - $x;
-        my $cyy  = $cy + $y;
-        my $cy_y = $cy - $y;
+        my $cxx  = int($cx + $x);
+        my $cx_x = int($cx - $x);
+        my $cyy  = int($cy + $y);
+        my $cy_y = int($cy - $y);
         if ($filled) {
-            if (int($cyy) <=> $old_cyy) {
+            if ($cyy <=> $old_cyy) {
                 $self->line({'x' => $cxx, 'y' => $cyy,'xx' => $cx_x, 'yy' => $cyy});
-                $old_cyy = int($cyy);
+                $old_cyy = $cyy;
             }
-            if ((int($cy_y) <=> $old_cy_y) && (int($cyy) <=> int($cy_y))) {
+            if (($cy_y <=> $old_cy_y) && ($cyy <=> $cy_y)) {
                 $self->line({'x' => $cx_x, 'y' => $cy_y,'xx' => $cxx, 'yy' => $cy_y});
-                $old_cy_y = int($cy_y);
+                $old_cy_y = $cy_y;
             }
         } else {
             $self->plot({'x' => $cxx,  'y' => $cyy,  'pixel_size' => $size});
@@ -1148,8 +1195,6 @@ sub ellipse {
             $XChange      += $TwoBSquare;
         } ## end if ((($EllipseError * ...)))
     } ## end while ($StoppingX >= $StoppingY)
-    $old_cyy      = 0;
-    $old_cy_y     = 0;
     $x            = 0;
     $y            = $YRadius;
     $XChange      = ($YRadius * $YRadius);
@@ -1159,18 +1204,18 @@ sub ellipse {
     $StoppingY    = $TwoASquare * $YRadius;
 
     while ($StoppingX <= $StoppingY) {
-        my $cxx  = $cx + $x;
-        my $cx_x = $cx - $x;
-        my $cyy  = $cy + $y;
-        my $cy_y = $cy - $y;
+        my $cxx  = int($cx + $x);
+        my $cx_x = int($cx - $x);
+        my $cyy  = int($cy + $y);
+        my $cy_y = int($cy - $y);
         if ($filled) {
-            if (int($cyy) <=> $old_cyy) {
+            if ($cyy <=> $old_cyy) {
                 $self->line({'x' => $cxx, 'y' => $cyy, 'xx' => $cx_x, 'yy' => $cyy});
-                $old_cyy = int($cyy);
+                $old_cyy = $cyy;
             }
-            if ((int($cy_y) <=> $old_cy_y) && (int($cyy) <=> int($cy_y))) {
+            if (($cy_y <=> $old_cy_y) && ($cyy <=> $cy_y)) {
                 $self->line({'x' => $cx_x, 'y' => $cy_y, 'xx' => $cxx, 'yy' => $cy_y});
-                $old_cy_y = int($cy_y);
+                $old_cy_y = $cy_y;
             }
         } else {
             $self->plot({'x' => $cxx,  'y' => $cyy,  'pixel_size' => $size});
@@ -1218,6 +1263,7 @@ sub circle {
     my $r      = $params->{'radius'};
     my $filled = $params->{'filled'}     || 0;
     my $size   = $params->{'pixel_size'} || 1;
+    
     $self->ellipse({'x' => $x, 'y' => $y, 'xradius' => $r, 'yradius' => $r, 'filled' => $filled, 'factor' => 1, 'pixel_size' => $size});
 } ## end sub circle
 
@@ -1309,10 +1355,10 @@ sub box {
     my $self   = shift;
     my $params = shift;
 
-    my $x      = $params->{'x'};
-    my $y      = $params->{'y'};
-    my $xx     = $params->{'xx'};
-    my $yy     = $params->{'yy'};
+    my $x      = int($params->{'x'});
+    my $y      = int($params->{'y'});
+    my $xx     = int($params->{'xx'});
+    my $yy     = int($params->{'yy'});
     my $filled = int($params->{'filled'} || 0);
     my $size   = int($params->{'pixel_size'} || 1);
     $size = 1 if ($filled);
@@ -1327,10 +1373,21 @@ sub box {
     if ($y > $yy) {
         ($y, $yy) = ($yy, $y);
     }
-    if ($filled == 1) {
+    if ($filled) {
         $w = abs($xx - $x);
         $h = abs($yy - $y);
-        $self->blit_write({'x' => $x, 'y' => $y, 'width' => $w, 'height' => $h, 'image' => $self->{'COLOR'} x ($w * $h)});
+        if (($self->{'accel_flags'} & $self->{'FBINFO_HWACCEL_FILLRECT'}) && $self->{'DRAW_MODE'} <= 1) {
+            _set_info(
+                $self->{'FBINFO_HWACCEL_FILLRECT'},
+                $self->{'FBinfo_hwaccel_fillrect'},
+                $self->{'FB'},
+                $x,$y,$w,$h,
+                $self->{'COLOR'},
+                $self->{'DRAW_MODE'}
+            );
+        } else {
+            $self->blit_write({'x' => $x, 'y' => $y, 'width' => $w, 'height' => $h, 'image' => $self->{'COLOR'} x ($w * $h)});
+        }
     } else {
         $self->polygon({'pixel_size' => $size, 'coordinates' => [$x, $y, $xx, $y, $xx, $yy, $x, $yy]});
     }
@@ -1675,7 +1732,16 @@ sub blit_copy {
     my $xx = int($params->{'x_dest'});
     my $yy = int($params->{'y_dest'});
 
-    $self->blit_write({'x' => $xx, 'y' => $yy, %{$self->blit_read({'x' => $x, 'y' => $y, 'width' => $w, 'height' => $h})}});
+    if (($self->{'accel_flags'} & $self->{'FBINFO_HWACCEL_COPYAREA'}) && $self->{'DRAW_MODE'} < 1) {
+        _set_info(
+            $self->{'FBINFO_HWACCEL_COPYAREA'},
+            $self->{'FBinfo_hwaccel_copyarea'},
+            $self->{'FB'},
+            $x,$y,$w,$h,$xx,$yy
+        );
+    } else {
+        $self->blit_write({'x' => $xx, 'y' => $yy, %{$self->blit_read({'x' => $x, 'y' => $y, 'width' => $w, 'height' => $h})}});
+    }
 } ## end sub blit_copy
 
 =head2 blit_read
@@ -1789,10 +1855,6 @@ sub blit_write {
     my $X_X = ($x + $self->{'XOFFSET'}) * $self->{'BYTES'};
     my ($index, $data, $px, $line, $idx, $px4);
 
-    #    if (   ($x >= $self->{'X_CLIP'})
-    #        && ($x <= $self->{'XX_CLIP'})
-    #        && ($y >= $self->{'Y_CLIP'})
-    #        && ($y <= $self->{'YY_CLIP'})) {
     if ($x < $self->{'X_CLIP'}) {
         $w += $x;
         $x = $self->{'X_CLIP'};
@@ -1928,8 +1990,6 @@ sub blit_write {
         } ## end for ($line = $y ; $line...)
     };
     $self->_fix_mapping() if ($@);
-
-    # } ## end if (($x >= $self->{'X_CLIP'...}))
 } ## end sub blit_write
 
 =head2 clip_reset
@@ -2093,10 +2153,10 @@ sub ttf_print {
     my $self   = shift;
     my $params = shift;
 
-    my $TTF_x       = $params->{'x'};
-    my $TTF_y       = $params->{'y'};
-    my $TTF_w       = $params->{'width'};
-    my $TTF_h       = $params->{'height'};
+    my $TTF_x       = int($params->{'x'});
+    my $TTF_y       = int($params->{'y'});
+    my $TTF_w       = int($params->{'width'});
+    my $TTF_h       = int($params->{'height'});
     my $P_color     = $params->{'color'};
     my $text        = $params->{'text'};
     my $face        = $params->{'face'};
@@ -2158,7 +2218,6 @@ sub ttf_print {
         );
         $self->blit_write({'x' => $TTF_x, 'y' => (($TTF_y - $TTF_h) + 1), 'width' => $TTF_w, 'height' => (($TTF_h + $global_ascent) - $global_descent), 'image' => $data});
     };
-#    print STDERR $@ if ($@);
     return (
         {
             'color'        => $params->{'color'},
@@ -2200,7 +2259,7 @@ sub load_image {
     my $self   = shift;
     my $params = shift;
 
-    my $img = Imager->new('channels' => 3);
+    my $img = Imager->new('channels' => 3); # Set up a 24 bit buffer
     return () unless ($img->read('file' => $params->{'file'}, 'allow_incomplete' => 0));
     my $orientation = $img->tags('name' => 'exif_orientation');
     if (defined($orientation) && $orientation) { # Automatically rotate the image
@@ -2220,13 +2279,16 @@ sub load_image {
         $img = $img->convert('matrix' => [[0, 0, 1], [0, 1, 0], [1, 0, 0]]);
     }
 
+    # Convert it to 32 bit
     $img = $img->convert('preset' => 'addalpha');
 
+    # Sometimes it works great, sometimes it looks uuuuuugly
     $img->filter('type' => 'autolevels') if ($params->{'autolevels'});
 
     my ($xs, $ys, $w, $h, %scale);
     $w = int($img->getwidth());
     $h = int($img->getheight());
+    # Scale the image, if asked to
     if ((defined($params->{'width'}) && $params->{'width'} <=> $w) || (defined($params->{'height'}) && $params->{'height'} <=> $h)) {
         $scale{'xpixels'} = $params->{'width'}  if (defined($params->{'width'}));
         $scale{'ypixels'} = $params->{'height'} if (defined($params->{'height'}));
@@ -2240,6 +2302,7 @@ sub load_image {
     $w = int($img->getwidth());
     $h = int($img->getheight());
     my $data = '';
+    # After all that mess, send the raw image data to $data
     $img->write(
         'type'          => 'raw',
         'interleave'    => 0,
@@ -2267,7 +2330,7 @@ sub load_image {
     $x = int($x);
     $y = int($y);
 
-    return (
+    return ( # return it in a form the blit routines can dig
         {
             'x'           => $x,
             'y'           => $y,
@@ -2364,7 +2427,7 @@ sub RGBA_to_16 {
     while ($big_data ne '') {
         my $pixel_data = substr($big_data, 0, 4);
         $big_data = substr($big_data, 4);
-        my ($b, $g, $r, $a) = unpack('I', $pixel_data);
+        my ($b, $g, $r, $a) = unpack('L', $pixel_data);
         $r = int($r / 8);
         $g = int($g / 8);
         $b = int($b / 8);
@@ -2396,7 +2459,7 @@ sub RGB_to_RGBA {
 
 sub _fix_mapping {
     my $self = shift;
-    unless($self->{'FILE_MODE'}) {
+    unless($self->{'FILE_MODE'}) { # Nothing to fix in file handle mode
         my $fb = $self->{'FB'};
         munmap($self->{'SCREEN'});
         mmap($self->{'SCREEN'}, $self->{'smem_len'}, PROT_READ | PROT_WRITE, MAP_SHARED, $fb);
@@ -2430,6 +2493,7 @@ sub _set_info {
     my $format  = shift;
     my $fb      = shift;
     my @array   = @_;
+
     my $data    = pack($format, @array);
     ioctl($fb, $command, $data);
 } ## end sub _set_info
@@ -2486,7 +2550,7 @@ modify it under the same terms as Perl itself.
 
 =head1 VERSION
 
-Version 4.22 (July 21, 2015)
+Version 4.23 (July 21, 2015)
 
 =head1 THANKS
 
